/**********************************************************************
  OrbitalExtension - Extension for generating cubes and meshes

  Copyright (C) 2008-2009 Marcus D. Hanwell

  This file is part of the Avogadro molecular editor project.
  For more information, see <http://avogadro.sourceforge.net/>

  Avogadro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Avogadro is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  02110-1301, USA.
 **********************************************************************/

#include "orbitalextension.h"

#include "gaussianfchk.h"
#include "slaterset.h"
#include "mopacaux.h"
#include "vdwsurface.h"

#include <vector>
#include <avogadro/toolgroup.h>
#include <avogadro/molecule.h>
#include <avogadro/atom.h>
#include <avogadro/cube.h>
#include <avogadro/mesh.h>
#include <avogadro/meshgenerator.h>
#include <Eigen/Core>

#include <QProgressDialog>
#include <QCoreApplication>
#include <QFileInfo>
#include <QMessageBox>
#include <QTime>
#include <QDebug>

using Eigen::Vector3d;
using Eigen::Vector3i;

namespace Avogadro
{
  OrbitalExtension::OrbitalExtension(QObject* parent) : Extension(parent),
    m_glwidget(0), m_orbitalDialog(0), m_molecule(0), m_basis(0), m_slater(0),
    m_progress(0), m_timer(0), m_meshGen1(0), m_meshGen2(0), m_VdWsurface(0)
  {
    QAction* action = new QAction(this);
    action->setText(tr("Create Surfaces..."));
    m_actions.append(action);
  }

  OrbitalExtension::~OrbitalExtension()
  {
    if (m_orbitalDialog) {
      delete m_orbitalDialog;
      m_orbitalDialog = 0;
    }
    if (m_basis) {
      delete m_basis;
      m_basis = 0;
    }
  }

  QList<QAction *> OrbitalExtension::actions() const
  {
    return m_actions;
  }

  QString OrbitalExtension::menuPath(QAction*) const
  {
    return tr("&Extensions");
  }

  QUndoCommand* OrbitalExtension::performAction(QAction *, GLWidget *widget)
  {
    m_glwidget = widget;
    if (!m_orbitalDialog) {
      m_orbitalDialog = new OrbitalDialog();
      m_orbitalDialog->setGLWidget(widget);
      m_orbitalDialog->setMolecule(m_molecule);
      connect(m_orbitalDialog, SIGNAL(calculateMO(int)),
              this, SLOT(calculateMO(int)));
      connect(m_orbitalDialog, SIGNAL(calculateAll()),
              this, SLOT(calculateAll()));
      connect(m_orbitalDialog, SIGNAL(calculateDensity()),
              this, SLOT(calculateDensity()));
      connect(m_orbitalDialog, SIGNAL(calculateMesh(int, double, int)),
              this, SLOT(generateMesh(int, double, int)));
      connect(m_orbitalDialog, SIGNAL(calculateVdWCube()),
              this, SLOT(calculateVdWCube()));
      connect(m_orbitalDialog, SIGNAL(calculateVdWMesh(int, double)),
              this, SLOT(generateVdWMesh(int, double)));
      setDefaultCube();
      if (loadBasis()) {
        m_orbitalDialog->show();
      }
      else {
        m_orbitalDialog->show();
      }
    }
    else {
      m_orbitalDialog->setGLWidget(widget);
      m_orbitalDialog->show();
    }
    return 0;
  }

  void OrbitalExtension::setMolecule(Molecule *molecule)
  {
    m_molecule = molecule;
    if (m_orbitalDialog)
      m_orbitalDialog->setMolecule(molecule);
  }

  void OrbitalExtension::setDefaultCube()
  {
    // Now to set the default cube...
    Cube cube;
    double step = 0.18;
    cube.setLimits(m_molecule, step, 2.5);
    Vector3d min = cube.min();// / BOHR_TO_ANGSTROM;
    Vector3i dim = cube.dimensions();
    // Set these values on the form - they can then be altered by the user
    m_orbitalDialog->setCube(min, dim.x(), dim.y(), dim.z(), step);
  }

  bool OrbitalExtension::loadBasis()
  {
    if (m_molecule->fileName().isEmpty()) {
      return false;
    }
    else if (m_loadedFileName == m_molecule->fileName()) {
      return true;
    }

    // Everything looks good, a new basis set needs to be loaded
    QFileInfo info(m_molecule->fileName());
    if (info.completeSuffix() == "fchk" || info.completeSuffix() == "fch") {
      if (m_basis)
        delete m_basis;
      m_basis = new BasisSet;
      GaussianFchk fchk(m_molecule->fileName(), m_basis);

      m_orbitalDialog->setMOs(m_basis->numMOs());
      for (int i = 0; i < m_basis->numMOs(); ++i) {
        if (m_basis->HOMO(i)) m_orbitalDialog->setHOMO(i);
        else if (m_basis->LUMO(i)) m_orbitalDialog->setLUMO(i);
      }
      return true;
    }
    else if (info.completeSuffix() == "mopout" || info.completeSuffix() == "out") {
      if (m_basis)
        delete m_basis;
      m_slater = new SlaterSet;
      MopacAux aux(info.absolutePath() + "/" + info.baseName() + ".aux", m_slater);

      // Set the number of MOs
      m_orbitalDialog->setMOs(m_slater->numMOs());
      for (unsigned int i = 0; i < m_slater->numMOs(); ++i) {
        if (m_slater->HOMO(i)) m_orbitalDialog->setHOMO(i);
        else if (m_slater->LUMO(i)) m_orbitalDialog->setLUMO(i);
      }
      return true;
    }
    // If we get here it is a basis set we cannot load yet
    else {
      qDebug() << "baseName:" << info.completeSuffix();
      m_orbitalDialog->setCurrentTab(0);
      return false;
    }
  }

  void OrbitalExtension::calculateMO(int mo, const Vector3d &origin,
                                     const Vector3i &steps, double stepSize)
  {
    const double BOHR_TO_ANGSTROM = 0.529177249;
    // Slater type orbital
    if (m_slater) {
      // We have a slater type orbital....
      qDebug() << "Adding a slater type orbital...." << mo;
      Cube *cube = m_molecule->addCube();
      cube->setName(QString(tr("MO ") + QString::number(mo)));
      cube->setLimits(origin * BOHR_TO_ANGSTROM, steps,
                      stepSize * BOHR_TO_ANGSTROM);
      if (!m_timer) {
        m_timer = new QTime;
        m_timer->start();
      }
      m_slater->calculateCubeMO(cube, mo);

      // Set up a progress dialog
      if (!m_progress) {
        m_progress = new QProgressDialog(m_orbitalDialog);
        m_progress->setCancelButtonText(tr("Abort Calculation"));
        m_progress->setWindowModality(Qt::NonModal);
      }

      // Set up the progress bar
      m_progress->setWindowTitle(tr("Calculating MO ") + QString::number(mo));
      m_progress->setRange(m_slater->watcher().progressMinimum(),
                           m_slater->watcher().progressMinimum());
      m_progress->setValue(m_slater->watcher().progressValue());
      m_progress->show();

      // Connect the signals and slots
      connect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
              m_progress, SLOT(setValue(int)));
      connect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
              m_progress, SLOT(setRange(int, int)));
      connect(m_progress, SIGNAL(canceled()),
              this, SLOT(slaterCanceled()));
      connect(&m_slater->watcher(), SIGNAL(finished()),
              this, SLOT(slaterDone()));
      m_orbitalDialog->enableCalculation(false);
    }
    else if (m_basis) {
      // Calculate each point in the cube in parallel
      Cube *cube = m_molecule->addCube();
      cube->setName(QString(tr("MO ") + QString::number(mo)));
      cube->setLimits(origin * BOHR_TO_ANGSTROM, steps,
                      stepSize * BOHR_TO_ANGSTROM);
      if (!m_timer) {
        m_timer = new QTime;
        m_timer->start();
      }
      m_basis->calculateCubeMO2(cube, mo);

     // Set up a progress dialog
      if (!m_progress) {
        m_progress = new QProgressDialog(m_orbitalDialog);
        m_progress->setCancelButtonText(tr("Abort Calculation"));
        m_progress->setWindowModality(Qt::NonModal);
      }

      // Set up the progress bar
      m_progress->setWindowTitle(tr("Calculating MO ") + QString::number(mo));
      m_progress->setRange(m_basis->watcher2().progressMinimum(),
                           m_basis->watcher2().progressMinimum());
      m_progress->setValue(m_basis->watcher2().progressValue());

      // Connect signals and slots
      connect(&m_basis->watcher2(), SIGNAL(progressValueChanged(int)),
              m_progress, SLOT(setValue(int)));
      connect(&m_basis->watcher2(), SIGNAL(progressRangeChanged(int, int)),
              m_progress, SLOT(setRange(int, int)));
      connect(m_progress, SIGNAL(canceled()),
              this, SLOT(calculation2Canceled()));
      connect(&m_basis->watcher2(), SIGNAL(finished()),
              this, SLOT(calculation2Done()));
      m_orbitalDialog->enableCalculation(false);
    }
  }

  void OrbitalExtension::calculateMO(int n)
  {
    if (!m_basis && !m_slater)
      return;

    const double BOHR_TO_ANGSTROM = 0.529177249;
    const double ANGSTROM_TO_BOHR = 1.0/BOHR_TO_ANGSTROM;
    // Calculate MO n and add the cube to the molecule...
    ++n; // MOs are 1 based, not 0 based...
    qDebug() << "Calculating MO" << n;
    double step = m_orbitalDialog->stepSize() * ANGSTROM_TO_BOHR;
    Vector3d origin = ANGSTROM_TO_BOHR * m_orbitalDialog->origin();
    Vector3i nSteps = m_orbitalDialog->steps();
    // Debug output
    qDebug() << "Origin = " << origin.x() << origin.y() << origin.z()
             << "\nStep = " << step << ", nz = " << nSteps.z();

    m_currentMO = 0; // Only calculating one MO
    calculateMO(n, origin, nSteps, step);
  }

  void OrbitalExtension::calculateAll()
  {
    if (!m_basis && !m_slater)
      return;

    const double BOHR_TO_ANGSTROM = 0.529177249;
    const double ANGSTROM_TO_BOHR = 1.0/BOHR_TO_ANGSTROM;
    qDebug() << "Calculating all MOs";
    m_stepSize = m_orbitalDialog->stepSize() * ANGSTROM_TO_BOHR;
    m_origin = ANGSTROM_TO_BOHR * m_orbitalDialog->origin();
    m_steps = m_orbitalDialog->steps();

    // Call the calculation, starting at MO 1
    m_currentMO = 1;
    calculateMO(m_currentMO, m_origin, m_steps, m_stepSize);
  }

  void OrbitalExtension::calculateDensity()
  {
    if (!m_basis && !m_slater)
      return;

    const double BOHR_TO_ANGSTROM = 0.529177249;
    const double ANGSTROM_TO_BOHR = 1.0/BOHR_TO_ANGSTROM;
    qDebug() << "Calculating electron density...";
    double stepSize = m_orbitalDialog->stepSize() * ANGSTROM_TO_BOHR;
    Vector3d origin = ANGSTROM_TO_BOHR * m_orbitalDialog->origin();
    Vector3i steps = m_orbitalDialog->steps();

    // Call the calculation, starting at MO 1
    m_currentMO = 0;
    if (m_slater) {
      // We have a slater type orbital....
      Cube *cube = m_molecule->addCube();
      cube->setName(QString(tr("Electron Density")));
      cube->setLimits(origin * BOHR_TO_ANGSTROM, steps,
                      stepSize * BOHR_TO_ANGSTROM);
      if (!m_timer) {
        m_timer = new QTime;
        m_timer->start();
      }
      m_slater->calculateCubeDensity(cube);

      // Set up a progress dialog
      if (!m_progress) {
        m_progress = new QProgressDialog(m_orbitalDialog);
        m_progress->setCancelButtonText(tr("Abort Calculation"));
        m_progress->setWindowModality(Qt::NonModal);
      }

      // Set up the progress bar
      m_progress->setWindowTitle(tr("Calculating Electron Density"));
      m_progress->setRange(m_slater->watcher().progressMinimum(),
                           m_slater->watcher().progressMinimum());
      m_progress->setValue(m_slater->watcher().progressValue());

      // Connect the signals and slots
      connect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
              m_progress, SLOT(setValue(int)));
      connect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
              m_progress, SLOT(setRange(int, int)));
      connect(m_progress, SIGNAL(canceled()),
              this, SLOT(slaterCanceled()));
      connect(&m_slater->watcher(), SIGNAL(finished()),
              this, SLOT(slaterDone()));
      m_orbitalDialog->enableCalculation(false);
    }
    else if (m_basis) {
      // FIXME Not implemented yet...
    }
  }

  void OrbitalExtension::calculationDone()
  {
    disconnect(&m_basis->watcher(), SIGNAL(progressValueChanged(int)),
               m_progress, SLOT(setValue(int)));
    disconnect(&m_basis->watcher(), SIGNAL(progressRangeChanged(int, int)),
            m_progress, SLOT(setRange(int, int)));
    disconnect(m_progress, SIGNAL(canceled()),
            this, SLOT(calculationCanceled()));
    disconnect(&m_basis->watcher(), SIGNAL(finished()),
            this, SLOT(calculationDone()));

    qDebug() << "Whole cube calculation done in" << m_timer->elapsed() / 1000.0
             << "seconds";
    delete m_timer;
    m_timer = 0;
    m_molecule->update();
    m_orbitalDialog->enableCalculation(true);
  }

  void OrbitalExtension::calculation2Done()
  {
    // Calculation complete
    if (!m_currentMO) {
      disconnect(&m_basis->watcher2(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_basis->watcher2(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
      disconnect(m_progress, SIGNAL(canceled()),
                 this, SLOT(calculation2Canceled()));
      disconnect(&m_basis->watcher2(), SIGNAL(finished()),
                 this, SLOT(calculation2Done()));

      qDebug() << "Single points calculation done in" << m_timer->elapsed() / 1000.0
               << "seconds";
      delete m_timer;
      m_timer = 0;
      m_molecule->update();
      m_orbitalDialog->enableCalculation(true);
    }
    else if (static_cast<unsigned int>(m_basis->numMOs()) == m_currentMO) {
      // All MOs have been calculated
      disconnect(&m_basis->watcher2(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_basis->watcher2(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
      disconnect(m_progress, SIGNAL(canceled()),
                 this, SLOT(calculation2Canceled()));
      disconnect(&m_basis->watcher2(), SIGNAL(finished()),
                 this, SLOT(calculation2Done()));

      qDebug() << "All cube MOs calculated in" << m_timer->elapsed() / 1000.0
               << "seconds";
      delete m_timer;
      m_timer = 0;
      m_molecule->update();
      m_orbitalDialog->enableCalculation(true);
      m_currentMO = 0;
    }
    else { // More work to do
      disconnect(&m_basis->watcher2(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_basis->watcher2(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
      disconnect(m_progress, SIGNAL(canceled()),
                 this, SLOT(calculation2Canceled()));
      disconnect(&m_basis->watcher2(), SIGNAL(finished()),
                 this, SLOT(calculation2Done()));
      calculateMO(++m_currentMO, m_origin, m_steps, m_stepSize);
    }
  }

  void OrbitalExtension::slaterDone()
  {
    if (!m_currentMO) {
      disconnect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
      disconnect(m_progress, SIGNAL(canceled()),
                 this, SLOT(slaterCanceled()));
      disconnect(&m_slater->watcher(), SIGNAL(finished()),
                 this, SLOT(slaterDone()));

      qDebug() << "Single points calculation done in" << m_timer->elapsed() / 1000.0
               << "seconds";
      delete m_timer;
      m_timer = 0;
      m_molecule->update();
      m_orbitalDialog->enableCalculation(true);
    }
    else if (m_slater->numMOs() == m_currentMO) {
      disconnect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
    disconnect(m_progress, SIGNAL(canceled()),
            this, SLOT(slaterCanceled()));
      disconnect(&m_slater->watcher(), SIGNAL(finished()),
                 this, SLOT(slaterDone()));

      qDebug() << "All cube MOs calculated in" << m_timer->elapsed() / 1000.0
               << "seconds";
      delete m_timer;
      m_timer = 0;
      m_molecule->update();
      m_orbitalDialog->enableCalculation(true);
      m_currentMO = 0;
    }
    else {
      disconnect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
                 m_progress, SLOT(setValue(int)));
      disconnect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
                 m_progress, SLOT(setRange(int, int)));
    disconnect(m_progress, SIGNAL(canceled()),
            this, SLOT(slaterCanceled()));
      disconnect(&m_slater->watcher(), SIGNAL(finished()),
                 this, SLOT(slaterDone()));
      calculateMO(++m_currentMO, m_origin, m_steps, m_stepSize);
    }
  }

  void OrbitalExtension::calculationCanceled()
  {
    disconnect(&m_basis->watcher(), SIGNAL(progressValueChanged(int)),
               m_progress, SLOT(setValue(int)));
    disconnect(&m_basis->watcher(), SIGNAL(progressRangeChanged(int, int)),
            m_progress, SLOT(setRange(int, int)));
    connect(m_progress, SIGNAL(canceled()),
            this, SLOT(calculationCanceled()));
    disconnect(&m_basis->watcher(), SIGNAL(finished()),
            this, SLOT(calculationDone()));
    m_basis->watcher().cancel();
    qDebug() << "Canceled...";
    m_progress->deleteLater();
    m_orbitalDialog->enableCalculation(true);
  }

  void OrbitalExtension::calculation2Canceled()
  {
    disconnect(&m_basis->watcher2(), SIGNAL(progressValueChanged(int)),
               m_progress, SLOT(setValue(int)));
    disconnect(&m_basis->watcher2(), SIGNAL(progressRangeChanged(int, int)),
            m_progress, SLOT(setRange(int, int)));
    connect(m_progress, SIGNAL(canceled()),
            this, SLOT(calculation2Canceled()));
    disconnect(&m_basis->watcher2(), SIGNAL(finished()),
            this, SLOT(calculation2Done()));
    m_basis->watcher2().cancel();
    qDebug() << "Canceled...";
    m_orbitalDialog->enableCalculation(true);
    m_currentMO = 0;
  }

  void OrbitalExtension::slaterCanceled()
  {
    disconnect(&m_slater->watcher(), SIGNAL(progressValueChanged(int)),
               m_progress, SLOT(setValue(int)));
    disconnect(&m_slater->watcher(), SIGNAL(progressRangeChanged(int, int)),
            m_progress, SLOT(setRange(int, int)));
    connect(m_progress, SIGNAL(canceled()),
            this, SLOT(slaterCanceled()));
    disconnect(&m_slater->watcher(), SIGNAL(finished()),
            this, SLOT(slaterDone()));
    m_slater->watcher().cancel();
    qDebug() << "Canceled...";
    m_orbitalDialog->enableCalculation(true);
    m_currentMO = 0;
  }

  void OrbitalExtension::generateMesh(int iCube, double isoValue, int)
  {
    if (!m_molecule->cube(iCube))
      return;

    Cube *cube = m_molecule->cube(iCube);
    m_mesh1 = m_molecule->addMesh();
    m_mesh1->setName(cube->name());
    m_mesh1->setIsoValue(isoValue);
    m_mesh1->setCube(cube->id());
    m_mesh2 = m_molecule->addMesh();
    m_mesh2->setName(cube->name());
    m_mesh2->setIsoValue(-isoValue);
    m_mesh2->setCube(cube->id());
    m_mesh1->setOtherMesh(m_mesh2->id());
    m_mesh2->setOtherMesh(m_mesh1->id());
    if (!m_meshGen1) {
      m_meshGen1 = new MeshGenerator;
      connect(m_meshGen1, SIGNAL(finished()), this, SLOT(meshGenerated()));
    }
    if (!m_meshGen2) {
      m_meshGen2 = new MeshGenerator;
      connect(m_meshGen2, SIGNAL(finished()), this, SLOT(meshGenerated()));
    }
    m_meshGen1->initialize(cube, m_mesh1, isoValue);
    m_meshGen1->start();
    m_meshGen2->initialize(cube, m_mesh2, -isoValue, true);
    m_meshGen2->start();
  }

  void OrbitalExtension::meshGenerated()
  {
    Engine *engine = m_orbitalDialog->currentEngine();
    QSettings settings;
    engine->writeSettings(settings);
    settings.setValue("mesh1Id", static_cast<int>(m_mesh1->id()));
    settings.setValue("mesh2Id", static_cast<int>(m_mesh2->id()));
    engine->readSettings(settings);
    engine->setEnabled(true);
    m_molecule->update();
  }

  void OrbitalExtension::calculateVdWCube()
  {
    if (!m_VdWsurface)
      m_VdWsurface = new VdWSurface;

    if (m_molecule)
      m_VdWsurface->setAtoms(m_molecule);
    else
      return;

    qDebug() << "Calculating VdW cube...";
    double stepSize = m_orbitalDialog->stepSize();
    Vector3d origin = m_orbitalDialog->origin();
    Vector3i steps = m_orbitalDialog->steps();
    Cube *cube = m_molecule->addCube();
    cube->setName(QString(tr("VdW Cube")));
    cube->setLimits(origin, steps, stepSize);

    m_VdWsurface->calculateCube(cube);

    // Set up a progress dialog
    if (!m_progress) {
      m_progress = new QProgressDialog(m_orbitalDialog);
      m_progress->setCancelButtonText(tr("Abort Calculation"));
      m_progress->setWindowModality(Qt::NonModal);
    }

    // Set up the progress bar
    m_progress->setWindowTitle(tr("Calculating VdW Cube"));
    m_progress->setRange(m_VdWsurface->watcher().progressMinimum(),
                         m_VdWsurface->watcher().progressMinimum());
    m_progress->setValue(m_VdWsurface->watcher().progressValue());

    connect(&m_VdWsurface->watcher(), SIGNAL(progressValueChanged(int)),
            m_progress, SLOT(setValue(int)));
    connect(&m_VdWsurface->watcher(), SIGNAL(progressRangeChanged(int, int)),
            m_progress, SLOT(setRange(int, int)));
    connect(m_progress, SIGNAL(canceled()),
            this, SLOT(calculateVdWCanceled()));
    connect(&m_VdWsurface->watcher(), SIGNAL(finished()),
            this, SLOT(calculateVdWDone()));
    m_progress->show();
  }

  void OrbitalExtension::calculateVdWDone()
  {
    disconnect(&m_VdWsurface->watcher(), 0, m_progress, 0);
    disconnect(m_progress, 0, this, 0);
  }

  void OrbitalExtension::calculateVdWCanceled()
  {
    disconnect(&m_VdWsurface->watcher(), 0, m_progress, 0);
    disconnect(m_progress, 0, this, 0);
    m_VdWsurface->watcher().cancel();
    qDebug() << "Canceled...";
  }

  void OrbitalExtension::generateVdWMesh(int iCube, double isoValue)
  {
    qDebug() << "Calculate VdW Mesh called" << isoValue << iCube;
    if (!m_molecule->cube(iCube))
      return;

    Cube *cube = m_molecule->cube(iCube);
    m_mesh1 = m_molecule->addMesh();
    m_mesh1->setName(cube->name());
    m_mesh1->setIsoValue(isoValue);
    m_mesh1->setCube(cube->id());

    if (!m_meshGen1) {
      m_meshGen1 = new MeshGenerator;
      connect(m_meshGen1, SIGNAL(finished()), this, SLOT(VdWMeshGenerated()));
    }
    else {
      disconnect(m_meshGen1, 0, this, 0);
      connect(m_meshGen1, SIGNAL(finished()), this, SLOT(VdWMeshGenerated()));
    }
    m_meshGen1->initialize(cube, m_mesh1, isoValue, true);
    m_meshGen1->start();
    qDebug() << "Calculate Mesh called" << isoValue;
  }

  void OrbitalExtension::VdWMeshGenerated()
  {
    Engine *engine = m_orbitalDialog->currentEngine();
    QSettings settings;
    engine->writeSettings(settings);
    settings.setValue("meshId", static_cast<int>(m_mesh1->id()));
    engine->readSettings(settings);
    engine->setEnabled(true);
    m_molecule->update();
  }

} // End namespace Avogadro

#include "orbitalextension.moc"

Q_EXPORT_PLUGIN2(orbitalextension, Avogadro::OrbitalExtensionFactory)
